/*
 * IO functions (io.c)
 * Part of libpogo, a c-library replacent for GBA
 * Programmed by Jonas Minnberg (Sasq)
 *
 * DESCRIPTION
 * Implements many of the standard c-library IO functions.
 *
 **/

int pogo_errno = 0;

#include "pogo/stdcore.h"
#include "pogo/device.h"
#include "pogo/io.h"

POGOFILE *pogo_stdin = POGO_NULL;
POGOFILE *pogo_stdout = POGO_NULL;
POGOFILE *pogo_stderr = POGO_NULL;

#ifdef POGO_PLATFORM_GBA
#define DEF_BUFSIZE 512
#define DEF_BUFCOUNT 2
#else
#define DEF_BUFSIZE 8192
#define DEF_BUFCOUNT 16
#endif

//int pogo_dprintf(char *fmt, ...);

typedef int POGO_FILE_HANDLE;

typedef struct
{
	int no;
	uchar *ptr;
}  BufInfo;

typedef struct
{
	int pos;
	int fileSize;
	POGO_FILE_HANDLE fd;
	int pageSize;
	int bufCount;
	BufInfo *buf;

} POGOFILE_T;

// "Touches" a buffer so it gets bumped to the top of the list
static void bump(POGOFILE_T *fp, int bno)
{
	int i;
	BufInfo tmp = fp->buf[bno];

	for(i=bno; i>0; i--)
		fp->buf[i] = fp->buf[i-1];

	fp->buf[0] = tmp;
}


// Get page 'no' in the file 'fp'.
static uchar *get_page(POGOFILE_T *fp, int no)
{
	int i;

	for(i=0; i<fp->bufCount; i++)
	{
		if(fp->buf[i].no == no)
		{
			bump(fp, i);
			//_dprintf("Cache hit (%d)\n", no);
			return fp->buf[0].ptr;
		}
	}

	bump(fp, fp->bufCount-1);

	fp->buf[0].no = no;
	pogo_lseek(fp->fd, fp->pageSize * no, PG_SEEK_SET);
	pogo_read(fp->fd, fp->buf[0].ptr, fp->pageSize);
	return fp->buf[0].ptr;

}

static int read_section(POGOFILE_T *fp, int from, int to, void *dest)
{
	int i,size;
	uchar *ptr, *d;
	int startp,endp;

	if(to >= fp->fileSize)
		to = fp->fileSize-1;

	if(from > to) return -1;

	// Cacluate start and stop page
	startp = from/fp->pageSize;
	endp = to/fp->pageSize;

	d = (uchar *)dest;

	for(i = startp; i<=endp; i++)
	{
		ptr = get_page(fp, i);
		size = fp->pageSize;
		if(i == startp)
		{
			ptr += (from % fp->pageSize);
			size = fp->pageSize - (from % fp->pageSize);
		}

		if(i == endp)
		{
			size = size - (fp->pageSize - (to % fp->pageSize) - 1);
		}

		memcpy(d, ptr, size);
		d += size;
	}

	return 1;
}


static POGOFILE_T *create_filep(int size, int bc, int ps)
{
	int i;
	POGOFILE_T *fp = (POGOFILE_T *)malloc(sizeof(POGOFILE_T));

	fp->bufCount = bc;
	fp->pageSize = ps;
	fp->pos = 0;
	fp->fileSize = size;

	if(bc)
	{
		fp->buf = (BufInfo *)malloc(fp->bufCount * sizeof(BufInfo));
		for(i=0; i<fp->bufCount; i++)
		{
			fp->buf[i].no = -1;
			fp->buf[i].ptr = malloc(fp->pageSize);
		}
	}
	else
		fp->buf = POGO_NULL;


	return fp;
}

static void delete_filep(POGOFILE_T *fp)
{
	int i;
	if(fp->bufCount)
	{
		for(i=0; i<fp->bufCount; i++)
			free(fp->buf[i].ptr);
		free(fp->buf);
	}

}

int pogo_ftell(POGOFILE *fp_in)
{
	POGOFILE_T *fp = (POGOFILE_T*)fp_in;
	if(!fp->bufCount)
		return pogo_tell(fp->fd);

	return fp->pos;
}

int pogo_feof(POGOFILE* fp_in)
{
	POGOFILE_T *fp = (POGOFILE_T*)fp_in;
	return (fp->pos >= fp->fileSize);
}

int pogo_fseek(POGOFILE *fp_in, int offset, int origin)
{
	POGOFILE_T *fp = (POGOFILE_T*)fp_in;

	if(!fp->bufCount)
		return pogo_lseek(fp->fd, offset, origin);


	switch(origin)
	{
	case PG_SEEK_SET:
		fp->pos = offset;
		break;
	case PG_SEEK_CUR:
		fp->pos += offset;
		break;
	case PG_SEEK_END:
		fp->pos = fp->fileSize - offset;
		break;
	}

	if(fp->pos < 0) fp->pos = 0;
	if(fp->pos > fp->fileSize) fp->pos = fp->fileSize;

	return fp->pos;
}

POGOFILE *pogo_fopen(const char *name, const char *mode)
{
	POGO_FILE_HANDLE fd;
	int m = 0;
	const char *p = mode;
	while(*p) {
		if(*p == 'w')
			m |= PGO_WRONLY|PGO_TRUNC;
		if(*p == 'r')
			m |= PGO_RDONLY;
		if(*p == '+')
		{
			if(m & PGO_RDONLY)
				m |= PGO_WRONLY;
			else
			if(m & PGO_WRONLY)
			{
				m &= ~PGO_TRUNC;
				m |= PGO_CREAT|PGO_RDONLY;
			}	
		}
		if(*p == 'a')
			m |= PGO_APPEND;
		p++;
	}

	//if((m & PGO_WRONLY) && !(m & PGO_APPEND))
	//	m |= (PGO_CREAT | PGO_TRUNC);

	fd = pogo_open(name, m);

	if(fd >= 0)
	{
		int bs = DEF_BUFSIZE;
		int bc = DEF_BUFCOUNT;

		int size;
		POGOFILE_T *fp;
		pogo_lseek(fd, 0, PG_SEEK_END);
		size = pogo_tell(fd);
		pogo_lseek(fd, 0, PG_SEEK_SET);
		//pogo_dprintf("Creating file size %d\n", size);
		//if(size > 65536)
		//	bs = 65536;
		if((size < 0) || !name || (m & PGO_WRONLY))
			fp = create_filep(size, 0, 0);
		else
			fp = create_filep(size, bc, bs);
		fp->fd = fd;
		return (POGOFILE *)fp;
	}
	else
		return POGO_NULL;
}

int pogo_fread(void *dest, int size, int count, POGOFILE *fp_in)
{
	int rc;
	POGOFILE_T *fp = (POGOFILE_T *)fp_in;
	if(!fp)
		return 0;

	size *= count;

	if(!fp->bufCount)
		return pogo_read(fp->fd, dest, size);

	if(fp->pos + size > fp->fileSize)
		size = fp->fileSize - fp->pos;

	rc = read_section(fp, fp->pos, fp->pos + size - 1, dest);

	fp->pos += size;

	if(rc < 0)
		return rc;
	else
		return size;

}


int pogo_fwrite(const void *src, int size, int count, POGOFILE *fp_in)
{
	int rc;
	POGOFILE_T *fp = (POGOFILE_T *)fp_in;
	if(!fp)
		return 0;
	rc = (int)pogo_write(fp->fd, src, size*count);
	if(rc > 0)
		fp->pos += rc;
	return rc;
}

int pogo_fclose(POGOFILE *fp_in)
{
	POGOFILE_T *fp = (POGOFILE_T *)fp_in;
	int rc = pogo_close(fp->fd);
	delete_filep(fp);
	return rc;
}

int pogo_fflush(POGOFILE_T *fp)
{
	return 0;
}

int pogo_fprintf(POGOFILE *fp, char *format, ...)
{
	char tmp[2560];
	int rc;
	va_list vl;
	va_start(vl, format);

	rc = vsprintf(tmp, format, vl);
	pogo_fwrite(tmp, 1, strlen(tmp), fp);

	va_end(vl);
	return rc;
}

int pogo_printf(char *format, ...)
{
	char tmp[2560];
	int rc;
	va_list vl;
	va_start(vl, format);

	rc = vsprintf(tmp, format, vl);
	pogo_write(POGO_STDOUT, tmp, strlen(tmp));

	va_end(vl);
	return rc;
}

int pogo_putc(int c, POGOFILE *fp)
{
	unsigned char cc = c;
	pogo_fwrite(&cc, 1, 1, fp);
	return c;
}


int pogo_putchar(int c)
{
	unsigned char cc = c;
	pogo_write(FD_STDOUT, &cc, 1);
	return c;
}

int pogo_getc(POGOFILE *fp)
{
	unsigned char c;
	if(pogo_fread(&c, 1, 1, fp) < 1)
		return POGO_EOF;
	else
		return c;
}

int pogo_fputc(int c, POGOFILE *fp)
{
	return pogo_putc(c, fp);
}

int pogo_fgetc(POGOFILE *fp)
{
	return pogo_getc(fp);
}

int pogo_ioctlv(int fd, int request, va_list vl);

int pogo_fioctl(POGOFILE *fp_in, int request, ...)
{
	int rc;
	va_list vl;
	POGOFILE_T *fp = (POGOFILE_T *)fp_in;

	va_start(vl, request);
	rc = pogo_ioctlv(fp->fd, request, vl);
	va_end(vl);
	return rc;
}


int pogo_getchar(void)
{
	unsigned char c;
	if(pogo_read(FD_STDIN, &c, 1) < 1)
		return POGO_EOF;
	else
		return c;
}

int pogo_fputs(char *string, POGOFILE *fp)
{
	return pogo_fwrite(string, 1, strlen(string), fp);
}

int pogo_puts(char *string )
{
	char lf = 10;
	pogo_write(FD_STDOUT, string, strlen(string));
	return pogo_write(FD_STDOUT, &lf, 1);
}

char *pogo_fgets(char *buffer, int n, POGOFILE *fp)
{
	int c = 0;
	char *ptr = buffer;
	n--;
	while((ptr-buffer < n) && c != 10)
	{
		if(pogo_fread(ptr, 1, 1, fp) == 1)
			c = *ptr++;
		else
			c = 10;
	}
	*ptr = 0;
	if(!strlen(buffer))
		return POGO_NULL;

	return buffer;
}

char *pogo_gets(char *buffer)
{
	int c = 0;
	char *ptr = buffer;

	while(c != 10)
	{
		while((c = pogo_getchar()) == POGO_EOF);
		if(c == 8)
		{
			if(ptr > buffer)
				ptr--;
		}
		else
		if(c != 10)
			*ptr++ = c;
	}
	*ptr = 0;
	return buffer;
}

static int readdir_rc;

struct POGODIR_T
{
	char name[128];
	struct pogo_dirent de;
	int fd;
};

POGODIR *pogo_opendir(const char *name)
{
	int rc = pogo_dopen(name);
	//pogo_fprintf(pogo_stderr, "dopen %s = %d\n", name, rc);
	if(rc >= 0)
	{
		struct POGODIR_T *d = malloc(sizeof(struct POGODIR_T));
		d->fd = rc;
		d->de.d_name = d->name;
		return (POGODIR *)d;
	}
	return POGO_NULL;
}


struct pogo_dirent *pogo_readdir(POGODIR *dir)
{
	struct POGODIR_T *d = (struct POGODIR_T*)dir;
	if(pogo_dread(d->fd, &d->name, 128) > 0)
		return &d->de;
	return POGO_NULL;
}

void pogo_closedir(POGODIR *dir)
{
	struct POGODIR_T *d = (struct POGODIR_T*)dir;
	pogo_dclose(d->fd);
	free(d);
}

int pogo_dir_error(void)
{
	return readdir_rc;
}
