

/* These are routines for handling the screen, either for a specific hardware (PSP, NDS, GBA etc)
 * or for a general framebuffer. The screen is always handled like 8bit paletted.
 * Define either GBA, DS, PSP, or PC - (and optionally FB8, FB16 or FB32) before compiling
 * Then call screen_set_target() to set your rendering target and format
 * Then call screen_setcolors() to set the current palette. For DIRECT modes this will change the
 * single palette. For 16 and 32 bit framebuffers it will just change the translation table
 */

#include "pogo/platform.h"

#if (defined POGO_PLATFORM_NDS)
  #include <nds.h>
#elif (defined POGO_PLATFORM_GBA)
  #include "pogo/gba_defs.h"
#elif (defined POGO_PLATFORM_PSP)
  #include "pogo/psp.h"
#elif (defined POGO_PLATFORM_PC)
#else
  #error "Unknown POGO_PLATFORM"
#endif

#include "pogo/stdcore.h"
#include "pogo/platform.h"
#include "pogo/screen.h"

// Current location and size of framebuffer
static int g_screenWidth;
static int g_screenHeight;
static Pixel *g_screenPtr;
static int g_screenStride;
static int g_screenVStride;


static int g_screenRotation;


//#ifdef POGO_PLATFORM_GBA
//
//typedef uchar Pixel;
//
//#else

//#undef DIRECT

//#endif


#ifdef DIRECT
#define COPY_PIXEL(dst, src) *(dst) = *(src)
#else
static Pixel colors[MAX_COLOR];
uint32 screen_color(int n)
{
	return colors[n];
}
#define COPY_PIXEL(dst, src) *(dst) = colors[*(src)]
#endif


#undef SCREEN_ROTATION
#define SCREEN_ROTATION 0
#include "screen_rot.h"
#undef SCREEN_ROTATION
#define SCREEN_ROTATION 1
#include "screen_rot.h"
#undef SCREEN_ROTATION
#define SCREEN_ROTATION 2
#include "screen_rot.h"
#undef SCREEN_ROTATION
#define SCREEN_ROTATION 3
#include "screen_rot.h"


int screen_setcolors(uint32 *pal, int start, int size)
{
	int i;
	unsigned int col;

	if((start+size) > MAX_COLOR)
		size = (MAX_COLOR - start);

	for(i=0; i<size; i++)
	{
		//col = pal[i];
		colors[i+start] = pal[i]; //UINT2PIXEL(col);
	}

	return size;
}

void (*screen_copy)(int x, int y, uchar *src, int width, int height, int sw, int solid);
void (*screen_set)(int x, int y, int width, int height, uint32 col);
Pixel* (*screen_pixel)(int x, int y);
void (*screen_scroll)(int x, int y, int width, int height, int lines);

void screen_get_target(Pixel **ptr, int *width, int *height, int *rotation)
{
	if(ptr)
		*ptr = g_screenPtr;
	if(width)
		*width = g_screenWidth;
	if(height)
		*height = g_screenHeight;
	if(rotation)
		*rotation = g_screenRotation;
}

void screen_set_rotation(int rotation)
{
	int flip;

	g_screenRotation = rotation;	
	flip = (g_screenRotation == ROT_TOPRIGHT) || (g_screenRotation == ROT_BOTTOMLEFT);
	
	if(flip)
	{
		g_screenHeight = g_screenStride;
		g_screenWidth = g_screenVStride;
	}
	else
	{
		g_screenWidth = g_screenStride;
		g_screenHeight = g_screenVStride;
	}

	
	switch(rotation)
	{
	case ROT_TOPLEFT:
		screen_set = screen_set_tl;
		screen_copy = screen_copy_tl;
		screen_pixel = screen_pixel_tl;
		screen_scroll = screen_scroll_tl;
		break;
	case ROT_TOPRIGHT:
		screen_set = screen_set_tr;
		screen_copy = screen_copy_tr;
		screen_pixel = screen_pixel_tr;
		screen_scroll = screen_scroll_tr;
		break;
	case ROT_BOTTOMRIGHT:
		screen_set = screen_set_br;
		screen_copy = screen_copy_br;
		screen_pixel = screen_pixel_br;
		screen_scroll = screen_scroll_br;
		break;
	case ROT_BOTTOMLEFT:
		screen_set = screen_set_bl;
		screen_copy = screen_copy_bl;
		screen_pixel = screen_pixel_bl;
		screen_scroll = screen_scroll_bl;
		break;
	}

}


void screen_set_target(Pixel *ptr, int hstride, int vstride)
{
	int i, flip, resize = 0;
	if(ptr)
		g_screenPtr = ptr;
	
	
	flip = (g_screenRotation == ROT_TOPRIGHT) || (g_screenRotation == ROT_BOTTOMLEFT);
	

	if(hstride)
		g_screenStride = hstride;
	
	if(vstride)
		g_screenVStride = vstride;

	
	if(flip)
	{
		g_screenHeight = g_screenStride;
		g_screenWidth = g_screenVStride;
	}
	else
	{
		g_screenWidth = g_screenStride;
		g_screenHeight = g_screenVStride;
	}
		
}

void screen_line(int x0, int y0, int x1, int y1, uint32 col, int size)
{
	int x, y;
		
	if(x0 == x1 && y0 == y1)
	{
		//ptr[x0+width*y0] ^= 0xffff;
		return;
	}
		
	if(ABS(y1-y0) > ABS(x1-x0))
	{
		int deltax;

		if(y0 > y1)
		{
			SWAP(y0,y1);
			SWAP(x0,x1);
		}
		
		deltax = (x1-x0)*256/(y1-y0);

		for(x=(x0<<8), y=y0; y<y1; y++)		
		{
			*(screen_pixel(x>>8, y)) = col;
			if(size > 1)
			{
				*(screen_pixel((x>>8)+1, y)) = col;
				*(screen_pixel((x>>8), y+1)) = col;
				*(screen_pixel((x>>8)+1, y+1)) = col;
			}
			x+=deltax;
		}
	}
	else	
	{
		int deltay;
		if(x0 > x1)
		{
			SWAP(y0,y1);
			SWAP(x0,x1);
		}
		
		deltay = (y1-y0)*256/(x1-x0);
		
		for(y=(y0<<8), x=x0; x<x1; x++)		
		{
			*(screen_pixel(x, y>>8)) = col;
			if(size > 1)
			{
				*(screen_pixel(x+1, (y>>8))) = col;
				*(screen_pixel(x, (y>>8)+1)) = col;
				*(screen_pixel(x+1, (y>>8)+1)) = col;
			}
			y+=deltay;
		}
	}	
}
