#include <pogo/device.h>

#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>

char rootdir[256];

PogoDevice native;

static FILE *open_files[32];
static int open_count = 0;

static char *rootend;

const char *fixname(const char *name)
{
	if(name[0] == '/')
		strcpy(rootend, name);
	else
	{
		*rootend = '/';
		strcpy(&rootend[1], name);
	}
	return rootdir;
}


#ifdef _WIN32_MSVC

struct dirent
{
	char *d_name;
};

typedef struct
{
	struct dirent de;
	struct _finddata_t fd;
	intptr_t handle;
} DIR;


DIR *opendir(char *name)
{
	char tmp[256];
	DIR *dir = (DIR*)malloc(sizeof(DIR));
	dir->de.d_name = NULL;
	sprintf(tmp, "%s\\*", name);
        //printf("Looking for %s\n", tmp);^M
	dir->handle = _findfirst(tmp, &dir->fd);
	return dir;
}

struct dirent *readdir(DIR *dir)
{
	intptr_t rc = dir->handle;
	if(dir->de.d_name)
		rc = _findnext(dir->handle, &dir->fd);
	if(rc == -1)
		return NULL;
	dir->de.d_name = dir->fd.name;
	return &dir->de;
}

void closedir(DIR *dir)
{
	_findclose(dir->handle);
	free(dir);
}

int native_mkdir(const char *name)
{
	const char *name2 = fixname(name);
	return _mkdir(name);	
}

#else

int native_mkdir(const char *name)
{
	const char *name2 = fixname(name);
#ifdef _WIN32
	return mkdir(name);
#else
	return mkdir(name, S_IRWXU);
#endif
}


#endif

int native_open(const char *name, int flags)
{
	FILE *f;
	int i;
	
	const char *name2 = fixname(name);
	
	if(flags & PGO_WRONLY)
		f = fopen(name2, "wb");
	else
		f = fopen(name2, "rb");
	if(f)
	{
		for(i=0; i<open_count; i++)
		{
			if(!open_files[i])
				break;
		}
		if(i>=open_count)
			i = open_count++;
		
		open_files[i] = f;
		return i;
	}
	return -1;
}

int native_read(int fd, void *dest, int size)
{
	FILE *f = open_files[fd];
	return fread(dest, 1, size, f);
}

int native_write(int fd, const void *dest, int size)
{
	FILE *f = open_files[fd];
	return fwrite(dest, 1, size, f);
}

void native_close(int fd)
{
	FILE *f = open_files[fd];
	fclose(f);
	open_files[fd] = NULL;
}

int native_lseek(int fd, int offset, int origin)
{
	FILE *f = open_files[fd];
	fseek(f, offset, origin);
	return ftell(f);
}

int native_tell(int fd)
{
	FILE *f = open_files[fd];
	return ftell(f);
}

int native_ioctl(int fd, int request, ...)
{
	return -1;
}

int native_remove(const char *name)
{
	return -1;
}


static DIR *opendirs[16];
static int dcount = 0;


int native_dopen(const char *name)
{
	const char *name2 = fixname(name);
	
	DIR *d = opendir(name2);
	if(d)
	{
		opendirs[dcount++] = d;
		return dcount-1;
	}
	return -1;
}

int native_dread(int fd, void *dest, int len)
{
	DIR *d = opendirs[fd];
	
	struct dirent *de = readdir(d);
	
	if(de)
	{
		strncpy(dest, de->d_name, len);
		return strlen(dest);
	}
	return -1;
}

void native_dclose(int fd)
{
	closedir(opendirs[fd]);
	opendirs[fd] = NULL;
	
	while(dcount && !opendirs[dcount-1])
		dcount--;
}

int native_stat(const char *name, struct pogo_stat *buffer)
{
	struct stat s;
	
	const char *name2 = fixname(name);
	
	int rc = stat(name2, &s);
	if(rc == 0)
	{
		buffer->pst_size = s.st_size;
		buffer->pst_atime = (uint32)s.st_atime;
		if(s.st_mode & S_IFDIR)
			buffer->pst_mode = PGS_IFDIR;
		else
			buffer->pst_mode = 0;
	}

	return rc;
}



void native_init()
{
	memset(&native, 0, sizeof(PogoDevice));
	native.open = native_open;
	native.read = native_read;
	native.write = native_write;
	native.close = native_close;
	native.lseek = native_lseek;
	native.stat = native_stat;
	native.dopen = native_dopen;
	native.dread = native_dread;
	native.dclose = native_dclose;
	native.mkdir = native_mkdir;
	
	getcwd(rootdir, sizeof(rootdir));
	
	rootend = &rootdir[strlen(rootdir)];

	pogo_device_register(&native, "/disk", NULL, -1);
}
